/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "Gradient.h"
#include "NotImplemented.h"

#include "CSSParser.h"
#include "GraphicsContext.h"
#include "GraphicsContextPlatformPrivate.h"

#include <pixman.h>

#include <fcntl.h>
#include <unistd.h>
#include <algorithm>

#include <wkgfx_GC_Rendering.h>

#if !PLATFORM(WIIU)
#include <stdio.h> // printf
#include "renderer/WiiURenderer.h"
extern WiiURenderer * g_MainRenderer;
#endif
#include <PlatformOS.h>
#include <PlatformMemory.h>
using namespace LibPlatform;

// --------------------------------------------------------------------------------------------------
/*
	Gradients & Memory

	Gradients are implemented as color ramp textures. These can, depending on the parameters be
	quite sizable.

	One would expect the users of the Gradient class to be aware of this, but this is not so!

	JS - being one of the main users via the Canvas API - for example will manage each gradient
	as a JS object and hence garbage collect it and all. BUT: it is totally unaware of the memory
	spent on the texture and hence will keep these "smal objects" around for ages and ages. So long
	in fact that we can serious run out of memory.

	JS' garbage collector does know about "extra memory costs" associated with objects, but the code
	controlling gradients does not provide an easy spot to integrate the suitable calls.

	Hence we don't fight this and implement some memory management of our own:

	We define a maximum amount of memory we want to allot to the task and keep track of all instances
	in use. If we exceed the limit, we start removing textures, using a simple LRU algorithm. Gradients
	hence check each time if they need to recompute their textures.

	This method avoid accumulation of too much memory use, BUT: it can not avoid temporary peaks!

	If too many gradients are accumulated while the GPU had no chance to use them (& our code had no chance
	to free up the memory afterwards) we can still produce ugly peaks in memory usage... or even an
	OOM. But... that would really be VERY much the user's job to avoid!

*/
// --------------------------------------------------------------------------------------------------

std::list<PlatformGradientWiiU*>	*PlatformGradientWiiU::m_gradients;			//!< List of gradients with current texture storage
size_t								PlatformGradientWiiU::m_memUsage = 0;		//!< Memory usage (without logic overhead) of currently stored gradient textures

const size_t kMaxGradientStorageSize = 4 << 20;									//!< Maximum amount of memory spent on gradient storage

// --------------------------------------------------------------------------------------------------

namespace WebCore {

// --------------------------------------------------------------------------------------------------
/**
	Destroy the platform specific gradient data as needed for this gradient
*/
void Gradient::platformDestroy()
{
#if PLATFORM(WIIU)
    if (m_gradient)
	{
		// pull this instance from the list
		PlatformGradientWiiU::m_gradients->remove(m_gradient);
		// get rid of the texture data
		if (m_gradient->m_texture)
		{
			PlatformGradientWiiU::m_memUsage -= GraphicsContextPlatformPrivate::getTextureDataSize(m_gradient->m_texture);
			// note: only CPU writes, no need to flush GPU caches
			GraphicsContextPlatformPrivate::destroyTextureData(m_gradient->m_texture);
		}
        delete m_gradient;
		m_gradient = NULL;
    }
#else

	if (m_gradient)
		DeleteOpenGLTexture(m_gradient->m_GLTextureID);
	m_gradient = 0;
#endif
}

// --------------------------------------------------------------------------------------------------

static inline unsigned int _ceilPow2(unsigned int x) { --x; x=x|(x>>1); x=x|(x>>2); x=x|(x>>4); x=x|(x>>8); x=x|(x>>16); return (x+1); }

// --------------------------------------------------------------------------------------------------
/**
	Retrieve the platform specific gradient data (a texture on WiiU) and generate it if needed
*/
PlatformGradient Gradient::platformGradient()
{
	// Do we already have it computed?
    if (m_gradient)
	{
#if PLATFORM(WIIU)
		// Got a texture, too? It might have been purged!
		if (m_gradient->m_texture)
		{
			// All well. Pop it to the front of the list & return it...
			if (PlatformGradientWiiU::m_gradients->front() != m_gradient)
			{
				PlatformGradientWiiU::m_gradients->erase(find(PlatformGradientWiiU::m_gradients->begin(), PlatformGradientWiiU::m_gradients->end(), m_gradient));
				PlatformGradientWiiU::m_gradients->push_front(m_gradient);
			}
			return m_gradient;
		}
#endif
		
	}

	// A new one is needed!
	
    pixman_image_t *src_img;
    pixman_point_fixed_t p0;
    pixman_point_fixed_t p1;
    pixman_gradient_stop_t *stops;
	
    p0.x = pixman_double_to_fixed(m_p0.x());
    p0.y = pixman_double_to_fixed(m_p0.y());
    p1.x = pixman_double_to_fixed(m_p1.x());
    p1.y = pixman_double_to_fixed(m_p1.y());

#if PLATFORM(WIIU)
    // Setup color stops for pixman...
    int num_stops = m_stops.size();
	pixman_gradient_stop_t*	pStopsMem;
	pixman_gradient_stop_t	stopsArray[16];
	if (num_stops > 16)
	{
		pStopsMem = (pixman_gradient_stop_t *)Memory::AllocFromDefaultHeap(num_stops * sizeof(pixman_gradient_stop_t));
		stops = pStopsMem;
	}
	else
	{
		pStopsMem = NULL;
		stops = stopsArray;
	}
		
	if (stops)
	{
		for (int i = 0; i < num_stops; ++i)
		{
			stops[i].x = pixman_double_to_fixed(m_stops[i].stop);
			stops[i].color.red   = 0xffff * m_stops[i].red;
			stops[i].color.green = 0xffff * m_stops[i].green;
			stops[i].color.blue  = 0xffff * m_stops[i].blue;
			stops[i].color.alpha = 0xffff * m_stops[i].alpha; 
		}
	
		int width, height;

//OPT: the upper, left corner of the gradient textures - depending on the user defined setup of the gradients - may be: EMPTY
//     on could introduce a UV offset and not store that area at all
//OPT: we store linear gradients vertically. this can/will introduce major overhead for long gradients! The minimum pitch for
//     a RGBA8 texture (linear) seems to be 256 bytes. Hence we end up with 252 bytes wasted per row!!
		if (m_radial)
		{
			src_img = pixman_image_create_radial_gradient(&p0, &p1, pixman_double_to_fixed(m_r0), pixman_double_to_fixed(m_r1), stops, num_stops);

			// Two circles describe start and end of the gradient. They are centered around their respecitive center points.
			// Hence adding in the radius and taking the bigger one, will give us the rectangle we need to render to.
			width = std::max(m_p0.x() + m_r0, m_p1.x() + m_r1);
			height = std::max(m_p0.y() + m_r0, m_p1.y() + m_r1);
		}
		else
		{
			src_img = pixman_image_create_linear_gradient(&p0, &p1, stops, num_stops);

			width = std::max(m_p0.x(), m_p1.x());
			height = std::max(m_p0.y(), m_p1.y());
		}

		if (width < 1)  width  = 1;
		if (height < 1) height = 1;
#if PLATFORM(WIIU)
        width = std::max(static_cast<float>(width), m_destinationRect.width());
        height = std::max(static_cast<float>(height), m_destinationRect.height());
#endif

		// We have a gradiant source image?
		if (src_img)
		{
			pixman_image_set_repeat(src_img, PIXMAN_REPEAT_PAD);
		
			// Ok. Create our own gradient instance...
			if (!m_gradient)
				m_gradient = new PlatformGradientWiiU();
			if (m_gradient)
			{
				// we'll need to repeat those textures, better make them pow2!
				width = _ceilPow2(width);
				height = _ceilPow2(height);

				// Initialize the texture to store the data in...
				m_gradient->m_texture = GraphicsContextPlatformPrivate::createTexture(width, height, GraphicsContextPlatformPrivate::TEX_FMT_RGBA8);
				if (m_gradient->m_texture)
				{
					// Tell pixman about the layout of the texture data...
					int pitch, twidth, theight;
					void *tdata = GraphicsContextPlatformPrivate::accessTextureBuffer(m_gradient->m_texture, twidth, theight, pitch);
					pixman_image_t *dest_img = pixman_image_create_bits(PIXMAN_r8g8b8a8, twidth, theight, (uint32_t *)tdata, pitch);
					if (dest_img)
					{
						// Make pixman put its data directly into the texture we prepared... (this will be alpha-pre-multiplied & stretched as needed (see: pow2 vs. real dimensions))
						pixman_image_composite(PIXMAN_OP_SRC, src_img, NULL, dest_img, 0, 0, 0, 0, 0, 0, twidth, theight);
					
						// Take care of the caches...
						GraphicsContextPlatformPrivate::flushCacheForTexture(m_gradient->m_texture);
						
						// No longer needed...
						pixman_image_unref(dest_img);
						pixman_image_unref(src_img);
						if (pStopsMem)
							Memory::FreeToDefaultHeap(pStopsMem);

#if PLATFORM(WIIU)
						// Add gradient to global list
						PlatformGradientWiiU::m_gradients->push_front(m_gradient);
						PlatformGradientWiiU::m_memUsage += GraphicsContextPlatformPrivate::getTextureDataSize(m_gradient->m_texture);

						// How much did we accumulate already? Purge if we need to! (but keep the current one of course)
						while((PlatformGradientWiiU::m_memUsage > kMaxGradientStorageSize) && (PlatformGradientWiiU::m_gradients->size() > 1))
						{
							PlatformGradientWiiU* pGradient = PlatformGradientWiiU::m_gradients->back();
							PlatformGradientWiiU::m_gradients->pop_back();

							PlatformGradientWiiU::m_memUsage -= GraphicsContextPlatformPrivate::getTextureDataSize(pGradient->m_texture);
							// note: only CPU writes, no need to flush GPU caches
							GraphicsContextPlatformPrivate::destroyTextureData(pGradient->m_texture);
							pGradient->m_texture = NULL;
						}
#endif
						return m_gradient;
					}

					GraphicsContextPlatformPrivate::destroyTextureData(m_gradient->m_texture);
				}
				Memory::FreeToDefaultHeap(m_gradient);
				m_gradient = NULL;
			}
			pixman_image_unref(src_img);
		}
		else
		{
			#ifndef NDEBUG
				OS::Report("*** Warning: gradient failed.\n");
			#endif
		}
		if (pStopsMem)
			Memory::FreeToDefaultHeap(pStopsMem);
	}
#else
	// Setup color stops for pixman...
	WiiURenderer * Renderer = g_MainRenderer;

	LOG_SIMPLE_RENDER_MSG("***Gradient::platformGradient()\n");

	//If there's no global renderer, we can't do anything
	if (!Renderer)
	{
		printf("ERROR: Gradient::platformGradient() called before global renderer was initialized.\n");
		return 0;
	}

	int num_stops = m_stops.size();
	pixman_gradient_stop_t*	pStopsMem;
	pixman_gradient_stop_t	stopsArray[16];
	if (num_stops > 16)
	{
		pStopsMem = (pixman_gradient_stop_t *)new char[num_stops * sizeof(pixman_gradient_stop_t)];
		stops = pStopsMem;
	}
	else
	{
		pStopsMem = NULL;
		stops = stopsArray;
	}

	if (stops)
	{
		for (int i = 0; i < num_stops; ++i)
		{
			stops[i].x = pixman_double_to_fixed(m_stops[i].stop);
			stops[i].color.red   = 0xffff * m_stops[i].red;
			stops[i].color.green = 0xffff * m_stops[i].green;
			stops[i].color.blue  = 0xffff * m_stops[i].blue;
			stops[i].color.alpha = 0xffff * m_stops[i].alpha; 
		}

		int width, height;
		int comp_width;

		if (m_radial)
		{
			src_img = pixman_image_create_radial_gradient(&p0, &p1, pixman_double_to_fixed(m_r0), pixman_double_to_fixed(m_r1), stops, num_stops);

			// Two circles describe start and end of the gradient. They are centered around their respecitive center points.
			// Hence adding in the radius and taking the bigger one, will give us the rectangle we need to render to.
			width = std::max(m_p0.x() + m_r0, m_p1.x() + m_r1);
			if (width < 1) width = 1;
			height = std::max(m_p0.y() + m_r0, m_p1.y() + m_r1);
			if (height < 1) height = 1;

			comp_width = width;
		}
		else
		{
			src_img = pixman_image_create_linear_gradient(&p0, &p1, stops, num_stops);

			width = std::max(m_p0.x(), m_p1.x());
			if (width < 1) width = 1;
			height = std::max(m_p0.y(), m_p1.y());
			if (height < 1) height = 1;
			//Q: why 10?
			comp_width =  width;
		}

		// We have a gradiant source image?
		if (src_img)
		{
			pixman_image_set_repeat(src_img, PIXMAN_REPEAT_PAD);

			// Ok. Create our own gradient instance...
			m_gradient = new tWebKitGLTextureInfo;
			if (m_gradient)
			{
				m_gradient->m_GLTextureID = GAMEPAD_UNDEFINED_ID;
				m_gradient->m_Width = comp_width;
				m_gradient->m_Height = height;

				// Initialize the texture to store the data in...
				char * tdata = new char[comp_width * height * 4];

				if (tdata)
				{
					memset(tdata, 0, comp_width * height * 4);

					//NOTE -- PIXMAN_r8g8b8a8 causes this to generate RGBA data in reverse byte order, making it actually
					//			ABGR. Why this is, I don't know, but we'll use PIXMAN_a8b8g8r8 to get RGBA data instead.

					// Tell pixman about the layout of the texture data...
					//pixman_image_t *dest_img = pixman_image_create_bits(PIXMAN_r8g8b8a8, comp_width, height, (uint32_t *)tdata, comp_width * 4);
					pixman_image_t *dest_img = pixman_image_create_bits(PIXMAN_a8b8g8r8, comp_width, height, (uint32_t *)tdata, comp_width * 4);

					if (dest_img)
					{
						// Make pixman put its data directly into the texture we prepared...
						pixman_image_composite(PIXMAN_OP_SRC, src_img, NULL, dest_img, 0, 0, 0, 0, 0, 0, comp_width, height);

						// Setup render environment
						GraphicsContextPlatformPrivate::beginDraw();

						//TODO -- Create the OpenGL texture here
						GAMEPADERROR Result;
						m_gradient->m_GLTextureID = CreateOpenGLTexture(Result, comp_width, height, (unsigned char*)tdata, true);

						if (Result)
							printf("ERROR: Gradient failed: %s\n", WiiURenderer::ErrorToString(Result));

						// No longer needed...
						pixman_image_unref(dest_img);
						pixman_image_unref(src_img);
						if (pStopsMem)
							delete [] pStopsMem;

						return m_gradient;
					}

					delete [] tdata;
				}
				delete m_gradient;
				m_gradient = NULL;
			}
			pixman_image_unref(src_img);
		}
		else
		{
			printf("WARNING: Gradient::platformGradient() failed.\n");
		}
	if (pStopsMem)
		delete [] pStopsMem;
	}
#endif
    return NULL; 
}


// --------------------------------------------------------------------------------------------------
/**
	Fill the given rectangle with the given gradient in the given context
*/
void Gradient::fill(GraphicsContext* context, const FloatRect& rect)
{
#if PLATFORM(WIIU)
    m_destinationRect = rect;
#endif
	PlatformGradient pGradient = platformGradient();

	if (pGradient)
		context->platformContext()->gradientFill(pGradient, rect);
}

} //namespace
